#ifndef THC_INC
#define THC_INC

#include "THCGeneral.h"
#include "THCAllocator.h"
#include "THCBlas.h"
#include "THCCachingAllocator.h"
#include "THCCachingHostAllocator.h"
#include "THCSleep.h"
#include "THCStorage.h"
#include "THCStorageCopy.h"
#include "THCStream.h"
#include "THCTensor.h"
#include "THCTensorCopy.h"
#include "THCTensorRandom.h"
#include "THCTensorMath.h"
#include "THCTensorConv.h"

#endif
