#ifndef TH_CUDA_TENSOR_COPY_INC
#define TH_CUDA_TENSOR_COPY_INC

#include "THCTensor.h"
#include "THCGeneral.h"
#include "THCHalf.h"
#include "THCStream.h"

#include "generic/THCTensorCopy.h"
#include "THCGenerateAllTypes.h"

#endif
